/*
LocoNet Handregler (Throttle)

by Philipp Gahtow, 
http://pgahtow.de/wiki/index.php?title=Arduino_Loconet_Throttle

u8glib
https://github.com/olikraus/u8glib/wiki/userreference
*/

#include <avr/pgmspace.h>
#include <U8glib.h>
#include <EEPROM.h>
#include "LocoNet.h"

//Address of the handheld controller as ID:
#define ID 0x20

lnMsg        *LnPacket;
LnBuf        LnTxBuffer;

//#define DEBUG         // uncomment for debug

//Arduino Pin
// 2:DO:SHIFT LED
// 3:DI:SHIFT2 key
// 4:DI:EC11 A
// 5:DI:EC11 B
// 6:DI:EC11 key
// 7:DO:LocoNET Txd
// 8:DI:LocoNET Rxd
// 9:DO:TRACK LED
//10:DO:OLED CS
//11:DO:OLED DC
//12:DO:OLED D1 (DATA)
//13:DO:OLED SCK 
//A0:AI:analog key
//A1:DI:SHIFT1 key
//A4:DO:OLED_SDATA
//A5:DO:OLED_SCLOCK

#define Led_Shift 2     // Shift LED
#define Shift2Key 3     // Shift2 Button
#define EnAPin 4        // Incremental A
#define EnBPin 5        // Incremental B
#define DirPin 6        // EC11 switch
#define LoconetTxd 7    // Loconet OUT
#define LoconetRxd 8    // Loconet IN
#define Led_Track 9     // Track LED
#define analogKey A0    // F0 - F4 button
#define Shift1Key A1    // Shift1 Button

int EnA = 1;
int EnB = 0;
#define Down -1
#define Up 1

#define THROWN 0
#define CLOSED 32

#define AdrMax 2024       // adjustable address
#define SpeedMax 127       // original 127 highest speed. (maybe use levels here -> FredI!)
//example(= 34): 0, 3, 5, 7, 11, 15, 19, 23, 27, 31, 35, 39, 43, 47, 51, 55, 59, 63, 67, 71, 75, 79, 83, 87, 91, 95, 99,103,107,111,115,119,123,127

U8GLIB_SH1106_128X64 u8g(13, 12, 10, 11); // (DO-clock,D1-data,CS,DC)new reset pull-up resistor , /* reset=*/ 13); 
//U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE);  // I2C / TWI  

LocoNetThrottleClass Throttle;
LocoNetClass Loconet;
        
uint32_t  LastThrottleTimerTick;
uint32_t  AnalogKeyTimerTick;
uint32_t  LastAdrTimerTick;
uint32_t  DispTimerTick;
uint32_t  TimerTick;     // Anything timer
uint32_t  BlinkTick;     // Flashing timer

word Adresse = 1;      // default Adresse 16-bit
byte Slot = 0xFF;        // act. Slot to addr. (None)
byte AdrH = 0;           // high share address
byte AdrL = 0;           // low share Adress

#define EEPROMOn true    //should the EEPROM be used?
//EEPROM Storage areas:
#define EEPROMAdrH 13    // High Address
#define EEPROMAdrL 14    // Low Address
#define EEPROMSlot 16    // Slot

byte PointDir = 0;

// Screen state definition
enum {
  IniSc = 0,
  IniEvSc,
  AddressSc,
  AddressPosSc,       // Digit setting for address setting
  AddressNumSc,
  AddressEvSc,        //　Address setting value setting
  SpeedSc,
  SpeedEvSc,
  EmgSc,
  EmgEvSc,
  FunctionSc,
  FunctionEvSc,
  PointSc,
  PointEvSc,          // Point   
} ScreenState;

// u8g drawing state definition
enum {
  Idle = 0,
  DrawingStart,
  Drawing,
  DrawingCompleted,
} u8gState;

enum {
  None = 0,
  DIR_single_click,
  DIR_double_click,
  DIR_long_click,
  SHIFT_single_click,
  SHIFT_double_click,
  SHIFT_long_click,
  SHIFT2_single_click,
  Func0,
  Func1,
  Func2,
  Func3,
  Func4,  
} KeyEvent;

// Button position when function key is pressed → function number conversion table
char FuncTable[5][2]={
// 0 1  point line selection
  {0,0},
  {1,5},
  {2,6},
  {3,7},
  {4,8}
};

// Button position when point key is pressed → point number conversion table
char PointTable[5][2]={
// 0 1  point line selection
  {1,6},
  {2,7},
  {3,8},
  {4,9},
  {5,10}
};
              
int PointAddress = 1;         // original uncomment
int PointAddressMAX = 999;    // original uncomment
char PointLineSel = 0;

char AddressPosSel = 3;
char NumSel = 0;      // Variable for digit setting

// char KeyEvent = 0;
char EncoderEvent = 0;
int spd=1;

int init_start = 1;
//--------------------------------------------------------------------------------------------------
// Initialisation
//--------------------------------------------------------------------------------------------------
void setup()
{
  pinMode(analogKey,INPUT);         // F0 - F4 buttons
  pinMode(Shift1Key,INPUT_PULLUP);  // SHIFT1 button
  pinMode(Shift2Key,INPUT_PULLUP);  // SHIFT2 button
  pinMode(DirPin,INPUT_PULLUP);     // EC11 switch
  pinMode(EnAPin,INPUT_PULLUP);     // EC11 A
  pinMode(EnBPin,INPUT_PULLUP);     // EC11 B

  pinMode(Led_Shift, OUTPUT);       // Shift LED
  pinMode(Led_Track, OUTPUT);       // Track LED
  digitalWrite(Led_Shift, HIGH);    // HIGH-OFF,LOW-ON
  digitalWrite(Led_Track, HIGH);    // HIGH-OFF,LOW-ON 
  
  LocoNet.init(LoconetTxd);   // initialize the LocoNet interface
                              // TxPin,PD7(D7)
  Throttle.init(0, 0, ID);    // create hand controls

  #if defined (DEBUG)
    Serial.begin(57600);
    Serial.println("");
    Serial.println(F("hello, LocoNET/DCC Throttle "));    // F() Puts data in FLASH memory
  #endif  
  
  ScreenState = IniSc;
  
  getOldSlot();    //Get last address and slot from EEPROM if available!
}

//--------------------------------------------------------------------------------------------------
// main loop
//--------------------------------------------------------------------------------------------------
void loop() {
    
  EncoderState();
  DirKeyState();
  ShiftKeyState();
  Shift2KeyState();
  FunctionKeyState();
  DisplayState();
  adrState(0);
  pointState(0);
  
  #if defined(DEBUG)
    debugstate();
  #endif
    
  // Check for any received LocoNet packets
  LnPacket = LocoNet.receive() ;
  if (LnPacket)  {
    //LocoNet.processSwitchSensorMessage(LnPacket);         // this function will call the specially named functions below... 
    if (!LocoNet.processSwitchSensorMessage(LnPacket))
      Throttle.processMessage(LnPacket) ;
      
    #if defined(DEBUG)
      Serial.println("LnPacket");
    #endif
  }
  
  if(isTime(&LastThrottleTimerTick, 100)) {       // It seems that Loco needs access every 100ms
    Throttle.process100msActions(); 
    if (Throttle.getState() != TH_ST_IN_USE){     //no Slot - LED off
      digitalWrite(Led_Track, HIGH);
    }else
      digitalWrite(Led_Track, LOW);               //LED on
    if (Throttle.getState() == TH_ST_SLOT_MOVE){
      digitalWrite(Led_Track, HIGH);              //LED off
    }
  }
}


// --------------------------------------------------------------------------------
void DisplayState(){
    static char buf[5];

    switch(ScreenState){
      case IniSc:
            TimerTick = millis();
            ScreenState = IniEvSc;
            break;
      case IniEvSc:
            if(u8gState==Idle)
              u8gState = DrawingStart;          // Display state: Start drawing
            if( millis()-TimerTick > 2000 ){    // 1000=1sec
              ScreenState = SpeedSc;            // original ScreenState = AddressEvSc;
            }
            break;
                    
      // Address setting
      case AddressSc :
            u8gState = DrawingStart;      // Display state: Start drawing
            ScreenState = AddressPosSc;
            break;

      case AddressPosSc :                   // Address digit setting
            if(u8gState==Idle)
              u8gState = DrawingStart;      // Display state: Start drawing

            if(EncoderEvent == Up){           // Up processing
              AddressPosSel++;
              EncoderEvent = None;
              if ( AddressPosSel >= 4 )
                AddressPosSel = 0;
              u8gState = DrawingStart;   // Display state: Start drawing
            
            }else if(EncoderEvent == Down){     // Down processing
              AddressPosSel--;
              EncoderEvent = None;
              if ( AddressPosSel < 0 )
                AddressPosSel = 3;
              u8gState = DrawingStart;    // Display state: Start drawing
            
            }else if(KeyEvent == DIR_single_click){  // Single click: decision
              KeyEvent = None;
              ScreenState = AddressNumSc;
            
            }else if(KeyEvent == DIR_double_click){  // Double-click: save address and Go to speed screen
              KeyEvent = None;
              if (Adresse >= AdrMax)
                Adresse = 1;
              AdrL = Adresse;        //Address in 2x 8 bit spokes for EEPROM backup
              AdrH = Adresse >> 8;
              saveEEPROMAddresse();
              adrState(1);
              ScreenState = SpeedSc;
            
            }else if((KeyEvent >= DIR_long_click)&&(KeyEvent <= Func4)){
              KeyEvent = None;
            }
            break;

      case AddressNumSc:
            if( millis()-TimerTick > 500 ){         // added
              KeyEvent = None;                      // added
              sprintf(buf,"%04d",Adresse);
              NumSel = buf[AddressPosSel] - 0x30;   // Subtract 0x30 from a character to get a number
              ScreenState = AddressEvSc;
            }
            break;

      case AddressEvSc :                    //　Address value setting
            if(u8gState==Idle)
              u8gState = DrawingStart;      // Display state: Start drawing
            
            if(EncoderEvent == Up){           // Up processing
              NumSel++;
              EncoderEvent = None;
              if ( NumSel >= 10 )
                NumSel = 0;
              buf[AddressPosSel] = NumSel + 0x30; // Add 0x30 from number and convert to character
              Adresse = atoi(buf);
              u8gState = DrawingStart;    // Display state: Start drawing
            
            }else if(EncoderEvent == Down){  // Down processing
              NumSel--;
              EncoderEvent = None;
              if ( NumSel < 0 )
                NumSel = 9;
              buf[AddressPosSel] = NumSel + 0x30; // Add 0x30 from number and convert to character
              Adresse = atoi(buf);
              u8gState = DrawingStart;    // Display state: Start drawing
            
            }else if(KeyEvent == DIR_single_click){
              KeyEvent = None;
              ScreenState = AddressPosSc;
            
            }else if(KeyEvent == DIR_double_click){   // Double-click: save address and Go to speed screen
              KeyEvent = None;
              if (Adresse >= AdrMax)
                Adresse = 1;
              AdrL = Adresse;        //Address in 2x 8 bit spokes for EEPROM backup
              AdrH = Adresse >> 8;
              saveEEPROMAddresse();
              adrState(1);
              ScreenState = SpeedSc;
            
            }else if((KeyEvent >= DIR_long_click)&&(KeyEvent <= Func4)){
              KeyEvent = None;
            }
            break;      
            
      // Speed setting
      case SpeedSc :                       
            if(init_start == 1) {
              Throttle.releaseAddress();              // Release Loco,(Keep Address,Slot stays Active)
              adrState(1);
              init_start = 0;
            }
            //if(u8gState==Idle)
              u8gState = DrawingStart;                // Display state: Start drawing
            ScreenState = SpeedEvSc;

            #if defined(DEBUG)
              Serial.println("LocoNET_Connect");
            #endif            
            break;
            
      case SpeedEvSc :                                // Menu Event processing
            ThrottleCont();                           // Throttle increase / phenomena processing
            
            if(KeyEvent == DIR_long_click){           // DIR long press
                //Throttle.releaseAddress();          // ori uncomment Release Loco,(Keep Address,Slot stays Active)
                Throttle.freeAddress(Adresse);        // Free Loco,(Clear Address, Slot not Active)
                Throttle.releaseAddress();            // Release Loco, (Keep Address,  Slot stays Active)
                //KeyEvent = None;
                //Slot = 0;                           // original comment
                TimerTick = millis();
                ScreenState = AddressNumSc;           // To address setting screen
                      
                #if defined(DEBUG)
                  Serial.println("Dis Connect");
                #endif
                
            }else if(KeyEvent == DIR_double_click){
                KeyEvent = None;
                if (Throttle.getSpeed() == 1) {                     // At speed 1, double click to switch between Forward and Reverse
                  Throttle.setDirection(!Throttle.getDirection());  // switch direction
                  u8gState = DrawingStart;                          // Display state: Start drawing
                        
                  #if defined(DEBUG)
                    Serial.println(Throttle.getDirection());
                  #endif
                }
            
            }else if(KeyEvent == DIR_single_click){                // Func = 0 DIR click state?
                KeyEvent = None;
                if (Throttle.getState() == TH_ST_IN_USE ) {   // Addr. Available?   
                  if (Throttle.getSpeed() > 1) {    
                    Throttle.setSpeed(1);             // EMERGENCY STOP (speed at stop = 1)
                    ScreenState = EmgSc;
                    u8gState = DrawingStart;          // Display state: Start drawing
                                              
                    #if defined(DEBUG)
                      Serial.println("EMG STOP");
                    #endif
                  } 
                }
                      
            }else if((KeyEvent == SHIFT_single_click) || (KeyEvent == SHIFT_long_click)){
                //ScreenState = FunctionSc;            // To function setting screen
                PointLineSel++;
                if(PointLineSel >= 2)
                  PointLineSel = 0;
                u8gState = DrawingStart;
                KeyEvent = None;             

            }else if(KeyEvent == SHIFT_double_click) {
                KeyEvent = None;
                                     
            }else if(KeyEvent == SHIFT2_single_click){
                ScreenState = PointSc;              // To point setting screen
                KeyEvent = None;             
            
            }else if(KeyEvent >= Func0 && KeyEvent <= Func4){
                if(PointLineSel != 0)
                  PointLineSel = 1;
                Throttle.setFunction(FuncTable[KeyEvent-Func0][PointLineSel], !Throttle.getFunction(FuncTable[KeyEvent-Func0][PointLineSel]));
                KeyEvent = None; 
            }
            
            break;

      // Very stop screen
      case EmgSc:
            u8gState = DrawingStart;      // Display state: Start drawing
            TimerTick = millis();
            ScreenState = EmgEvSc;
            break;            

      case EmgEvSc:
            if(u8gState==Idle)
              u8gState = DrawingStart;          // Display state: Start drawing
            if( millis()-TimerTick > 1000 ){    // 1000=1sec
              ScreenState = SpeedSc;
            }
            break;
            

      // Function setting screen
      case FunctionSc:
            u8gState = DrawingStart;      // Display state: Start drawing
            ScreenState = FunctionEvSc;
            break;

      case FunctionEvSc:
            if(u8gState==Idle)
              u8gState = DrawingStart;      // Display state: Start drawing

            ThrottleCont();                 // Throttle increase / phenomena processing

            if((KeyEvent == DIR_single_click) || (KeyEvent == DIR_long_click)){ // Func = 0 DIR click state?
                KeyEvent = None;
                if (Throttle.getState() == TH_ST_IN_USE ) {   // Addr. Available?   
                    if (Throttle.getSpeed() > 1) {    
                      #if defined(DEBUG)
                        Serial.println("EMG STOP");
                      #endif
                      Throttle.setSpeed(1);             // EMERGENCY STOP (speed at stop = 1)
                      ScreenState = EmgSc;
                      u8gState = DrawingStart;          // Display state: Start drawing
                    }else {
                      ScreenState = SpeedSc;
                    }
                }else {
                    ScreenState = SpeedSc;  
                }
            
            }else if((KeyEvent == SHIFT_long_click)||(KeyEvent == SHIFT_double_click)) {   // to speed screen
                KeyEvent = None;
                ScreenState = SpeedSc;
            
            }else if(KeyEvent == SHIFT_single_click){// SHIFT Single click Change point line
                PointLineSel++;
                if(PointLineSel >= 2)     // ori 4
                  PointLineSel = 0;
                KeyEvent = None;
                u8gState = DrawingStart;
            
            }else if(KeyEvent == SHIFT2_single_click){// SHIFT2 Single click change point line
                if(PointLineSel > 0)
                  PointLineSel--;
                KeyEvent = None;
                u8gState = DrawingStart;
            
            }else if(KeyEvent == DIR_double_click){
                KeyEvent = None;
            
            }else if(KeyEvent >= Func0 && KeyEvent <= Func4){
                Throttle.setFunction(FuncTable[KeyEvent-Func0][PointLineSel], !Throttle.getFunction(FuncTable[KeyEvent-Func0][PointLineSel]) ); //Since FuncKey is incremented by 1, it is decremented by 1 when it is set to Loco
                                                                            //Should the control form be the same as Point, where the screen display does not match?
                #if defined(DEBUG)
                  Serial.print("FuncKey:");
                  Serial.print(KeyEvent-Func0, DEC);
                  Serial.print(" PointLineSel:");
                  Serial.println(PointLineSel, DEC);
                #endif
                KeyEvent = None;
            }
            break;      


      // Point setting screen
      case PointSc :
            u8gState = DrawingStart;      // Display state: Start drawing
            ScreenState = PointEvSc;
            break;

      case PointEvSc:
            if(u8gState==Idle)
              u8gState = DrawingStart;        // Display state: Start drawing
            //ThrottleCont();                 // original comment -- Throttle increase / phenomena processing

            if(EncoderEvent == Up){           // Up processing
                PointAddress++;
                EncoderEvent = None;
                if ( PointAddress >= PointAddressMAX )
                  PointAddress = PointAddressMAX;
                u8gState = DrawingStart;    // Display state: Start drawing
            
            }else if(EncoderEvent == Down){  // Down processing
                PointAddress--;
                EncoderEvent = None;
                if ( PointAddress <= 0 )
                  PointAddress = 1;
                u8gState = DrawingStart;    // Display state: Start drawing
            
            }else if((KeyEvent == DIR_single_click) || (KeyEvent == DIR_long_click)){   
                KeyEvent = None;
                if (Throttle.getState() == TH_ST_IN_USE ) {   // Addr. Available?   
                  if (Throttle.getSpeed() > 1) {    
                    #if defined(DEBUG)
                      Serial.println("EMG STOP");
                    #endif
                    Throttle.setSpeed(1);             // EMERGENCY STOP (speed at stop = 1)
                    ScreenState = EmgSc;
                    u8gState = DrawingStart;          // Display state: Start drawing
                  }else {
                    ScreenState = SpeedSc;
                  }
                }else {
                  ScreenState = SpeedSc;  
                }
            
            }else if(KeyEvent == SHIFT_single_click){// SHIFT Single click Change point line
                  PointLineSel++;
                  if(PointLineSel >= 4)       // ori 4
                    PointLineSel = 0;
                  KeyEvent = None;
                  u8gState = DrawingStart;
            
            }else if((KeyEvent == SHIFT_double_click)||(KeyEvent == SHIFT_long_click)) {  // to speed screen
                  KeyEvent = None;
                  ScreenState = SpeedSc;
            
            }else if(KeyEvent == DIR_double_click){
                  KeyEvent = None;
            
            }else if(KeyEvent == SHIFT2_single_click){// SHIFT2 Single click switch point
                  pointState(PointAddress);
                  KeyEvent = None;
            
            }else if(KeyEvent >= Func0 && KeyEvent <= Func4){
                  pointState(PointTable[KeyEvent-Func0][PointLineSel]);

                  #if defined(DEBUG)
                    Serial.print("FuncKey:");
                    Serial.print(KeyEvent-Func0, DEC);
                    Serial.print(" PointLineSel:");
                    Serial.println(PointLineSel, DEC);
                  #endif
                  KeyEvent = None;
            }
            break;      
      default:
            break;
    }
//----------------------------------------------------------------------------
  if(u8gState!=0){
    if( millis()-DispTimerTick > 100 ){    // 1000=1sec
      switch(u8gState){
        case DrawingStart:
              u8g.firstPage();               // Declaration to write on LCD
              u8gState = Drawing;            // Drawing
              break;

        case Drawing:
              do {
                ScreenDisplay();
              } while(u8g.nextPage()!=0);
              u8gState = DrawingCompleted;   // Finished drawing
              break;

        case DrawingCompleted:
              u8gState = Idle; 
              break;

        default :
              break;
      }
      DispTimerTick = millis();                 // Update to current time
    }
  }
}
  
// --------------------------------------------------------------------------------
// Display screen creation
// The screen to be displayed is determined by ScreenState
// --------------------------------------------------------------------------------
void ScreenDisplay(){

  char buf[22];
  char i;
  static char bf = 0;

  
  switch(ScreenState){
    case IniSc:
            break;
    case IniEvSc:
            u8g.setFont(u8g_font_profont12);
            u8g.drawStr( 0, 12*1, "Loco/DCC Throttle");
            u8g.drawStr( 0, 12*2, "Mode:[LocoNET]");
            sprintf(buf,"Initial Address:%04d",Adresse);
            u8g.drawStr( 0, 12*3, buf);
            u8g.drawStr( 0, 12*4, "Address:[variable]"); // fixedly: fixed, variable
            u8g.drawStr( 0, 12*5, "Version:2/2020");
            break;
            
    case AddressPosSc:  
    case AddressEvSc:
            u8g.setFont(u8g_font_profont12);
            u8g.drawStr( 0, 22, "Addr:");
            if(Throttle.getDirection())
              u8g.drawStr( 0, 48, "[<<--]");   // original - u8g.drawStr( 0, 46, "[FWD]");
            else
              u8g.drawStr( 0, 48, "[-->>]");   // original - u8g.drawStr( 0, 46, "[REV]");         
            u8g.drawStr( 0, 62, "Speed:");

            u8g.drawStr( 100, 62, "Km/h");

            u8g.setFont(u8g_font_gdr25r);
            sprintf(buf,"%04d",Adresse);
//               u8g.drawStr( 45, 24, buf);

            if( millis()-BlinkTick > 250 ){    // 1000=1sec
              if(bf == 0){
//              u8g.drawBox(40, 0, 80, 24);       // If you rewrite here, the lower half is not displayed?
//              u8g.setDefaultBackgroundColor();
//              u8g.drawStr( 40, 24, buf);
//              u8g.setDefaultForegroundColor();
                BlinkTick = millis();
                bf = 1;
              }else{
//              u8g.drawStr( 40, 24, buf);
                BlinkTick = millis();
                bf = 0;                
              }
            }

            if(bf==0){
                //u8g.drawBox(40, 25, 20, 2);
                if(ScreenState==AddressPosSc) cursor(40,25,0,AddressPosSel,buf);
                if(ScreenState==AddressEvSc){
                  cursor(40,25,1,AddressPosSel,buf);
                }
            } else{
             u8g.drawStr( 40, 24, buf); 
            }
            
            sprintf(buf,"%3d",spd-1);
            u8g.drawStr( 40, 63, buf);    
            break;

    case EmgSc:
    case EmgEvSc:
    case SpeedEvSc:
            u8g.setFont(u8g_font_profont12);
           
            u8g.drawStr( 0, 22, "Addr:");
            if(Throttle.getDirection())
              u8g.drawStr( 0, 48, "[<<--]");  // original - u8g.drawStr( 0, 46, "[FWD]");
            else
              u8g.drawStr( 0, 48, "[-->>]");  // original - u8g.drawStr( 0, 46, "[REV]");
            u8g.drawStr( 0, 62, "Speed:");

            if(PointLineSel == 0) {
              u8g.drawStr( 36, 36, "F0 F1 F2 F3 F4");
            }else {
              u8g.drawStr( 36, 36, "F0 F5 F6 F7 F8");
            }

            if(ScreenState == SpeedEvSc)
              u8g.drawStr( 100, 62, "Km/h");

            u8g.setFont(u8g_font_gdr25r);
            sprintf(buf,"%04d",Adresse);
            u8g.drawStr( 40, 24, buf);

            if(ScreenState == EmgSc || ScreenState == EmgEvSc)
              strcpy(buf, "EMG");
            else {
              sprintf(buf,"%3d",spd - 1);   // Throttle.getSpeed () gets 1-126, subtract -1
            }
            u8g.drawStr( 40, 63, buf);     
            break;

    case FunctionSc:
    case FunctionEvSc:
            FunctionCursor();
            break;

    case PointSc:
    case PointEvSc:
            PointCursor();
            
            break;
    default:
            break;             
    }
}
//---------------------------------------------------------------------------

void ThrottleCont(){
  spd = Throttle.getSpeed();
            
  if(EncoderEvent == Up){                   // Up processing
    if(KeyEvent == SHIFT_long_click){       // Press and hold SHIFT +5
      spd = spd + 5;                        
      KeyEvent == None;
    }else {
      if (spd <= 5) {
        spd++ ;                             // original +1
      }else if (spd > 5 && spd <= 20) {
        spd = spd + 2;
      }else if (spd >20 && spd <=50) {
        spd = spd + 3;
      }else if (spd >50 && spd <=SpeedMax) {
        spd = spd + 4;
      }
    }
    if ( spd >= SpeedMax )
      spd = SpeedMax;
           
    Throttle.setSpeed(spd);
    KeyEvent == None;
    EncoderEvent = None;
    u8gState = DrawingStart;          // Display state: Start drawing

  }else if(EncoderEvent == Down){      // Down processing
    if(KeyEvent == SHIFT_long_click){   // SHIFT long press -5
      KeyEvent == None;
      spd = spd - 5;
    }else {
      if (spd >50 && spd <=SpeedMax) {
        spd = spd - 4;
      }else if (spd >20 && spd <=50) {
        spd = spd - 3;
      }else if (spd >5 && spd <=20) {
        spd = spd - 2;
      }else if (spd <=5) {
        spd = spd - 1;              // original -1
      }
    }
    EncoderEvent = None;

    if ( spd < 1 )
      spd = 1;
    Throttle.setSpeed(spd);
    u8gState = DrawingStart;          // Display state: Start drawing
  }
}
                
//--------------------------------------------------------------------------------------------------
// FunctionCursor
// Cursor drawing on function setting screen
//--------------------------------------------------------------------------------------------------
void FunctionCursor(){
  char buf[20];
  u8g.setFont(u8g_font_profont12);
  sprintf(buf,"ADR:%04d  SPEED:%03d",Adresse, Throttle.getSpeed()-1);
  u8g.drawStr( 0, 12, buf);
  u8g.drawStr( 12,24,"F0  F1  F2  F3  F4");
  u8g.drawStr( 12,36,"F0  F5  F6  F7  F8");   
  u8g.drawStr( 12,48,"F0  F9  F10 F11 F12");   
  u8g.drawStr( 12,60,"F0  F13 F14 F15 F16");   
  u8g.drawStr( 0,24+12*PointLineSel,">");
}



//--------------------------------------------------------------------------------------------------
// Pointcursor
// Cursor drawing on point setting screen
//--------------------------------------------------------------------------------------------------
void PointCursor(){
  char buf[20];
  u8g.setFont(u8g_font_profont12);
  sprintf(buf,"PADR:%03d  SPEED:%03d",PointAddress, Throttle.getSpeed()-1);
  //sprintf(buf,"PointAddress:%03d",PointAddress);  // original
  u8g.drawStr( 0, 12, buf);
  u8g.drawStr( 12,24,"P1  P2  P3  P4  P5");
  u8g.drawStr( 12,36,"P6  P7  P8  P9  P10");   
  u8g.drawStr( 12,48,"P11 P12 P13 P14 P15");   
  u8g.drawStr( 12,60,"P16 P17 P18 P19 P20");  
  u8g.drawStr( 0,24+12*PointLineSel,">");
}

//--------------------------------------------------------------------------------------------------
// cursor
// Cursor blink processing
// mode 0: underline type
// mode 1: 1 character selection type
//--------------------------------------------------------------------------------------------------
void cursor(char x,char y, char mode,char pos, char *buf){
  char pbuf[5];

  switch(mode){
    case 0:
            u8g.drawStr( 40, 24, buf);
            u8g.drawBox( 40+pos*19, y, 20, 2);  
            break;
    case 1:
            for(int i=0;i<=4;i++){
              pbuf[0]=buf[i];pbuf[1]=0x00;
              if(i==pos){ // drawBox -> setDefaultBackgroundColor -> drawStr -> setDefaultForegroundColor // This order is important
                u8g.drawBox( 40+pos*19, y-25, 20, 25);
                u8g.setDefaultBackgroundColor();
                u8g.drawStr( 40+i*19, 24, pbuf);
                u8g.setDefaultForegroundColor();    
              } else
                u8g.drawStr( 40+i*19, 24, pbuf);
            }
            break;
    default:
            break;
  }
}



//--------------------------------------------------------------------------------------------------
// Encoder processing state
//--------------------------------------------------------------------------------------------------
void EncoderState(){
    int EnAState = digitalRead(EnAPin);

    EnB = digitalRead(EnBPin);
    
    if ((EnA == LOW) && (EnAState == HIGH)) {
      if (EnB == LOW) {
        EncoderEvent = Down;       // original Down
      } 
      else
        EncoderEvent = Up;     // original Up
    }
    
    EnA = EnAState;
}


//--------------------------------------------------------------------------------------------------
// SHIFT key processing state
// judge single click, double click, long press
//--------------------------------------------------------------------------------------------------
void ShiftKeyState(){
  static int state = 0;
  static char bufp = 0;
  static uint32_t SaveMillisTime;
  static uint32_t SaveMillisTime2;
  static char cKeybuf[12];
  
  char changes = 0;
  char SumBuf = 0;
  char i;
  char Shift1KeyRead;
  char Max = 10;

  switch(state){
    case 0:
            if (KeyEvent == None) {
              SaveMillisTime = millis();
              state = 1;
            }
            break;
    case 1:
            if( millis()- SaveMillisTime >100){              // original 50msec - Monitoring of 100msec cycle
              Shift1KeyRead = digitalRead(Shift1Key);        // Read button
              if(Shift1KeyRead == 0 || bufp > 0) {           // First detection? bufp is 0 or more
                cKeybuf[bufp++] = Shift1KeyRead;
                if(bufp >= Max){            
                  state = 2;
                  bufp = 0;
                  SaveMillisTime2 = millis();
                  break;
                }
              }
              SaveMillisTime = millis();
            }
            break;
    case 2:            // Single click / double click / long press judgment
            #if defined(DEBUG)
              Serial.println("");
              Serial.print("Keybuf:");  
              for(i=0;i<=Max-1;i++)
                Serial.print(cKeybuf[i],DEC);
              Serial.print(":");
            #endif                  
            for( i = 0 ; i <= Max - 2 ; i++){   // Change point is counted by variable changes
              if(cKeybuf[i] != cKeybuf[i+1])
                changes = changes + 1;   // number of changes
              SumBuf = SumBuf + cKeybuf[i];      // the sum of the buffer, If there are many 0, it can be used for long press judgment
            }
            #if defined(DEBUG)
              Serial.print(changes,DEC);
              Serial.print(":");   
              Serial.print(SumBuf,DEC);
              Serial.print(":");  
            #endif            
            switch(changes){
              case 0:
                     #if defined(DEBUG)
                       Serial.println("Long press");
                     #endif
                     KeyEvent = SHIFT_long_click;
                     digitalWrite(Led_Shift, LOW);  // HIGH-OFF,LOW-ON
                     break;
              case 1:
              case 2:
                      if(SumBuf<=5){
                         #if defined(DEBUG) 
                           Serial.println("Long press");
                         #endif
                         KeyEvent = SHIFT_long_click;
                         digitalWrite(Led_Shift, LOW);  // HIGH-OFF,LOW-ON
                      } else {
                        #if defined(DEBUG)
                          Serial.println("single click");
                        #endif
                        KeyEvent = SHIFT_single_click;
                        digitalWrite(Led_Shift, LOW);  // HIGH-OFF,LOW-ON
                      }
                      break;
              case 3:
              case 4:
                      #if defined(DEBUG)
                        Serial.println("Double click");
                      #endif
                      KeyEvent = SHIFT_double_click;
                      digitalWrite(Led_Shift, LOW);  // HIGH-OFF,LOW-ON
                     break;
              default:
                    break;
            }
            #if defined(DEBUG)
            // Serial.println("");
            #endif
            state = 3;
            break;
    case 3:        
            if( millis()-SaveMillisTime2 >1000) {
              digitalWrite(Led_Shift, HIGH);  // HIGH-OFF,LOW-ON
              state = 0;
            }
            break;
    default:
            break;
  }
}

//--------------------------------------------------------------------------------------------------
void Shift2KeyState(){
  static uint32_t SaveMillisTime;
  static int state = 0;
  
  if (state == 0) {
    if(digitalRead(Shift2Key) == 0) {
      SaveMillisTime = millis();
      digitalWrite(Led_Shift, LOW);  // HIGH-OFF,LOW-ON
      state = 1;
    }
    
  }else if (state == 1) {
    if(millis()-SaveMillisTime >500) {
      KeyEvent = SHIFT2_single_click;
      state = 2;
    }
  }else if (state == 2) {
    if(millis()-SaveMillisTime >1000) {
      digitalWrite(Led_Shift, HIGH);  // HIGH-OFF,LOW-ON
      state = 0;
    }
  }
}
//--------------------------------------------------------------------------------------------------
// Processing state of DIR key
// judge single click, double click, long press
//--------------------------------------------------------------------------------------------------
void DirKeyState(){
  static char state = 0;
  static char bufp = 0;
  static uint32_t SaveMillisTime;
  static char cKeybuf[12];
  
  char changes = 0;
  char SumBuf = 0;
  char i;
  char DirPinRead;
  char Max = 10;

  switch(state){
    case 0:
            if (KeyEvent == None) {
              SaveMillisTime = millis();
              state = 1;
            }
            break;
    case 1:                                       
            if( millis()-SaveMillisTime >100){          // original 50msec - Monitoring of 100msec cycle
              DirPinRead = digitalRead(DirPin);         // read button
              if(DirPinRead == 0 || bufp > 0) {         // First detection? bufp is 0 or more
                cKeybuf[bufp++] = DirPinRead;
                if(bufp >= Max){            
                  state = 2;
                  bufp = 0;
                  break;
                }
              }
              SaveMillisTime = millis();
            }
            break;
    case 2:                       // Single click / double click / long press judgment
            #if defined(DEBUG)
              Serial.println("");
              Serial.print("Keybuf:");  
              for(i=0;i<=Max-1;i++)
                Serial.print(cKeybuf[i],DEC);
              Serial.print(":");
            #endif                  
            for( i = 0 ; i <= Max - 2 ; i++){   // Change point is counted by variable changes
              if(cKeybuf[i] != cKeybuf[i+1])
                changes = changes + 1;          // number of changes
              SumBuf = SumBuf + cKeybuf[i];             // the sum of the buffer, If there are many 0, it can be used for long press judgment
            }
            #if defined(DEBUG)
              Serial.print(changes,DEC);
              Serial.print(":");   
              Serial.print(SumBuf,DEC);
              Serial.print(":");  
            #endif            
            switch(changes){
              case 0:
                      #if defined(DEBUG)
                       //Serial.println("Long press");
                      #endif
                      KeyEvent = DIR_long_click;
                      break;
              case 1:
              case 2:
                      if(SumBuf<=5){
                        #if defined(DEBUG)
                          //Serial.println("Long press");
                        #endif
                        KeyEvent = DIR_long_click;
                      } else {
                        #if defined(DEBUG)
                          //Serial.println("single click");
                        #endif
                        KeyEvent = DIR_single_click;
                      }
                      break;
              case 3:
              case 4:
                      #if defined(DEBUG)
                        Serial.println("Double click");
                      #endif
                      KeyEvent = DIR_double_click;
                      break;
              default:
                      break;
            }
            #if defined(DEBUG)
              Serial.println("");
            #endif
            state = 0;
            break;
    default:
            break;
  }  
}


//--------------------------------------------------------------------------------------------------
// Function key processing state
// Key judgment processing from analog keyboard
// The threshold value is adjusted separately, and the detection voltage is different when the initial detection value is continuously pressed.
// judge single click
// Add a 1 second filter because it will be imported continuously
//--------------------------------------------------------------------------------------------------
void FunctionKeyState(){
  static char state = 0;
  static int kin;
  int row;

  if(KeyEvent >= Func0 && KeyEvent <= Func4 )   // Unprocessed if analog key processing is not completed
    return;

  switch(state){
    case 0:
            kin = analogRead(analogKey);
            if( kin > 800 )                   // original 890 - Func button is no longer pressed?
              return;
  
            if( kin < 90){                   // ori-100 / F0,F0,F0 read value analog input 3 
              KeyEvent = Func0;
            } else if( kin < 220){           // ori-300 / F1,F5,F9 read value analog input 184
              KeyEvent = Func1;
            } else  if( kin < 380){          // ori-500 / F2,F6,F10 read value analog input 363
              KeyEvent = Func2;
            } else  if( kin < 550){          // ori-600 / F3,F7,F11 read value analog input 516
              KeyEvent = Func3;
            } else  if( kin < 750){          // ori-700 / F4,F8,F12 read value analog input 645
              KeyEvent = Func4;
            }
            AnalogKeyTimerTick = millis(); 
            state = 1;
  
            #if defined(DEBUG)          
              Serial.print("F:");
              Serial.print(kin,DEC);
              Serial.print(" KeyEvent:");
              Serial.println(KeyEvent,DEC);
            #endif
            
            break;
    case 1:
            if(millis()-AnalogKeyTimerTick > 500){  // 1000=1sec
              AnalogKeyTimerTick = millis();
              state = 0;
            }  
            break;
  }
}

//--------------------------------------------------------------------------------------------------
// Point switching state machine
// Changed after inquiring the current status. By doing so, it is convenient to switch c / t with one button
//--------------------------------------------------------------------------------------------------
void pointState(int adr){
  static int state = 0;
  static int pointAdr;

  #if defined(DEBUG)
    Serial.print("pointState:");
    Serial.println(state,DEC);
  #endif
  
  switch(state){
    case 0:
            if( adr != 0 ){
              pointAdr = adr;
              state = 1;
            }
            break;
    case 1:                                             // Get pointer status
            Loconet.reportSwitch(pointAdr);
            LastAdrTimerTick = millis();
            state = 2;
            break;
    case 2:                                             // Get address after 100ms
            if(millis()-LastAdrTimerTick>=100) {
              if(PointDir == THROWN)
                Loconet.requestSwitch(pointAdr,1,CLOSED);
              else
                Loconet.requestSwitch(pointAdr,1,THROWN);
              state = 0;
            }
            break;
    default:
            break;
  }
}

//--------------------------------------------------------------------------------------------------
// State machine for address acquisition
// Even if you execute freeAddress ()-> setAddress (), busy cannot be detected and processing cannot be performed.
//--------------------------------------------------------------------------------------------------

void adrState( int inh ){
  static int state = 0;

  #if defined(DEBUG)
    Serial.print("adrState:");
    Serial.println(state,DEC);
  #endif

  switch(state){
    case 0:
            if( inh == 1 )
              state = 1;
            break;
    case 1:                                             // release the acquired slot
            Throttle.freeAddress(Adresse);              // Free Loco,(Clear Address, Slot not Active)
            LastAdrTimerTick = millis();
            state = 2;
            break;
    case 2:                                             // Get address after 100ms
            if(millis()-LastAdrTimerTick>=100) {
              Throttle.setAddress(Adresse);             // Select Loco Address
              state = 0;
            }
            break;
    default:
            break;
  }
}

//--------------------------------------------------------------------------------------------------
void saveEEPROMAddresse() {
  if ((AdrL != EEPROM.read(EEPROMAdrL)) && (EEPROMOn == true))
      EEPROM.write(EEPROMAdrL, AdrL);
  if ((AdrH != EEPROM.read(EEPROMAdrH)) && (EEPROMOn == true))
      EEPROM.write(EEPROMAdrH, AdrH);
  if ((Slot != EEPROM.read(EEPROMSlot)) && (EEPROMOn == true))
      EEPROM.write(EEPROMSlot, Slot);
}  
//--------------------------------------------------------------------------------------------------
void getEEPROMAdresse() {
  byte AdrH = EEPROM.read(EEPROMAdrH);
  byte AdrL = EEPROM.read(EEPROMAdrL);
  Adresse = 0;
  Adresse = AdrH << 8;
  Adresse = Adresse + AdrL;
}

//--------------------------------------------------------------------------------------------------
//Get last slot from EEPROM if active
void getOldSlot() {
  if (Slot == 0xFF && EEPROMOn == true) {    //Check for reset?
    byte oldslot = EEPROM.read(EEPROMSlot);
    getEEPROMAdresse();
    if (oldslot == 255 || oldslot == 0)
      Slot = 0;   //no slot available to join
    else {
      Slot = oldslot;  
      Throttle.resumeAddress(Adresse, oldslot);
    }
  }
  else {
    if (Slot == 0xFF || Slot == 0)    //Slot not reset?
      Slot = 0;
    else Throttle.resumeAddress(Adresse, Slot);
  }  
}

//--------------------------------------------------------------------------------------------------
// update Speed via tick
// boolean type is a type that represents two values of true / false
//--------------------------------------------------------------------------------------------------
boolean isTime(unsigned long *timeMark, unsigned long timeInterval) {
    unsigned long timeNow = millis();
    if ( timeNow - *timeMark >= timeInterval) {
        *timeMark = timeNow;
        return true;
    }    
    return false;
}


//--------------------------------------------------------------------------------------------------
// Throttle notify Call-back functions
// Fly by if (function name), how does it work?
// Loconet message returns to LocoSlot-> DC50K-> LocoSlot? When you come back, are you updating?
// DCS50K STOP 1, FULL 126
//--------------------------------------------------------------------------------------------------
void notifyThrottleSpeed( uint8_t UserData, TH_STATE State, uint8_t Speed ) {

  #if defined(DEBUG)
  Serial.print("notifyThrottleSpeed():");
  Serial.println(Speed);
  #endif

  spd = Throttle.getSpeed();  // To update the OLED screen
  u8gState = DrawingStart;
}

//--------------------------------------------------------------------------------------------------
// Address is the vehicle address? Slot is the slot number where the vehicle address is stored
//--------------------------------------------------------------------------------------------------
void notifyThrottleAddress( uint8_t UserData, TH_STATE State, uint16_t Adresse, uint8_t isSlot) {

  #if defined(DEBUG)
  Serial.print("notifyThrottleAddress():adress");
  Serial.print(Adresse);
  Serial.print("_slot:");
  Serial.println(isSlot);
  #endif

  //updateLCDAdr(Adresse);
  Slot = isSlot;
}

//--------------------------------------------------------------------------------------------------
// forward or reverse?
//--------------------------------------------------------------------------------------------------
void notifyThrottleDirection( uint8_t UserData, TH_STATE State, uint8_t Direction )
{
  #if defined(DEBUG)
  Serial.print("notifyThrottleDirection():");
  Serial.println(Direction);
  #endif
}

//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
void notifyThrottleFunction( uint8_t UserData, uint8_t Function, uint8_t Value ) {

  #if defined(DEBUG)
    //Serial.print("notifyThrottleFunction():");
    //Serial.print(Function);
    //Serial.print(":");
    //Serial.println(Value);
    //return;
 
    if (Function == 0) {        // F0 reception?
      Serial.println("F0");
    }else {
      Serial.print("F");
      Serial.print(Function);
      Serial.print(":");
      Serial.println(Value);
    }
  #endif
}

//--------------------------------------------------------------------------------------------------
void notifyThrottleSlotStatus( uint8_t UserData, uint8_t Status ) {};

//--------------------------------------------------------------------------------------------------
// What do you see?
//--------------------------------------------------------------------------------------------------
void notifyThrottleState( uint8_t UserData, TH_STATE PrevState, TH_STATE State ) {
//  Serial.print("notifyThrottleState():");
//  Serial.println(Throttle.getStateStr(State));
}

//--------------------------------------------------------------------------------------------------
void notifyThrottleError( uint8_t UserData, TH_ERROR Error ) {
    #if defined(DEBUG)
//  Serial.print("notifyThrottleError():");
//  Serial.println(Throttle.getErrorStr(Error));
    #endif
}

//---------------------------------------------------------------------------------------------------
void notifySwitchRequest( uint16_t Adresse, uint8_t outData, uint8_t dirData) {

  #if defined(DEBUG)
  Serial.print("notifySwitchRequest():");
  Serial.print(Adresse,DEC);
  Serial.print(",");
  Serial.print(outData,DEC);   // Always 16?
  Serial.print(",");
  Serial.print(dirData,DEC);   // 0: t: Branch 32: c: Straight line
  Serial.println("");
  #endif
  
  PointDir = dirData;
}

//--------------------------------------------------------------------------------------------------
#if defined(DEBUG)
void debugstate(){
     if (Serial.available()) {
        // Process serial input for commands from the host.
        int ch = Serial.read();
        Serial.println(ch);
        if(ch=='A'){
          Serial.print("ADDRESS:");
          Serial.println(Adresse);          
        }
        if(ch=='S'){
          Serial.print("STATE:");
          Serial.print(Throttle.getState());
          Serial.print(":");
          Serial.println(Throttle.getStateStr(Throttle.getState()));
        }
        if(ch=='T'){
          Serial.print("Throttle.getSpeed:");
          Serial.println(Throttle.getSpeed());
        }

        if(ch=='I'){  // make addres free
          Serial.println("setAddress");      
          Throttle.setAddress(Adresse);         // Select Loco Address
        }
        if(ch=='Z'){  // free Address
          Serial.println("freeAddress");      
          Throttle.freeAddress(Adresse);        // Free Loco, (Clear Address, Slot not Active)
        }
        if(ch=='R'){  // release Address
          Serial.println("releaseAddress");      
          Throttle.releaseAddress();            // Release Loco, (Keep Address,  Slot stays Active)
        }
        if(ch=='O'){  // get addres
          Throttle.dispatchAddress(Adresse);    // Dispatch Loco Address
          Throttle.freeAddress(Adresse);        // Free Loco, (Clear Address, Slot not Active)
          Throttle.releaseAddress();            // Release Loco, (Keep Address, Slot stays Active)
        }
        if(ch=='D'){
          Throttle.acquireAddress();           // Acquire previously Dispatched Loco Address
        }

     }
}
#endif
